//
// pvtiff.c
//
// Write a "baseline" BMP to an uncompressed baseline TIFF file.
// Does not use any external libraries.
//
// I recommend against trying to modify this code to do anything
// more advanced than that.
//

#include <windows.h>
#include <tchar.h>
#include <stdio.h>

#include "purpleview.h"
#include <strsafe.h>

//	TIFF_NOTYPE	= 0,	/* placeholder */
#define TIFF_BYTE      1 /* 8-bit unsigned integer */
#define TIFF_ASCII     2 /* 8-bit bytes w/ last byte null */
#define TIFF_SHORT     3 /* 16-bit unsigned integer */
#define TIFF_LONG      4 /* 32-bit unsigned integer */
#define TIFF_RATIONAL  5 /* 64-bit unsigned fraction */
//	TIFF_SBYTE	= 6,	/* !8-bit signed integer */
//	TIFF_UNDEFINED	= 7,	/* !8-bit untyped data */
//	TIFF_SSHORT	= 8,	/* !16-bit signed integer */
//	TIFF_SLONG	= 9,	/* !32-bit signed integer */
//	TIFF_SRATIONAL	= 10,	/* !64-bit signed fraction */
//	TIFF_FLOAT	= 11,	/* !32-bit IEEE floating point */
//	TIFF_DOUBLE	= 12	/* !64-bit IEEE floating point */



// This flag determines the byte order of the TIFF file that
// will be written. It has nothing to do with the native byte
// order of the host computer. (This code is not particularly
// efficient, but is endian-agnostic.)
static const int little_endian = 1;

static void write_uint8(FILE *f,unsigned int x)
{
	unsigned int buf[1];
	buf[0]=(unsigned char)( x&0x000000ff    );
	fwrite(buf,1,1,f);
}


static void write_uint16(FILE *f,unsigned int x)
{
	unsigned char buf[2];
	if(little_endian) {
		buf[0]=(unsigned char)( x&0x000000ff    );
		buf[1]=(unsigned char)((x&0x0000ff00)>>8);
	}
	else {
		buf[0]=(unsigned char)((x&0x0000ff00)>>8);
		buf[1]=(unsigned char)( x&0x000000ff    );
	}
	fwrite(buf,1,2,f);
}

static void write_uint32(FILE *f,unsigned int x)
{
	unsigned char buf[4];
	if(little_endian) {
		buf[0]=(unsigned char)( x&0x000000ff    );
		buf[1]=(unsigned char)((x&0x0000ff00)>>8);
		buf[2]=(unsigned char)((x&0x00ff0000)>>16);
		buf[3]=(unsigned char)((x&0xff000000)>>24);
	}
	else {
		buf[0]=(unsigned char)((x&0xff000000)>>24);
		buf[1]=(unsigned char)((x&0x00ff0000)>>16);
		buf[2]=(unsigned char)((x&0x0000ff00)>>8);
		buf[3]=(unsigned char)( x&0x000000ff    );
	}
	fwrite(buf,1,4,f);
}

static void write_bytes(FILE *f,char *buf, int n)
{
	fwrite(buf,1,n,f);
}

int pv_tiff_save(HWND hwnd, struct iinfo_struct *iinfo,
				 struct einfo_struct *einfo)
{
	LPBITMAPINFOHEADER lpdib;
	unsigned char *lpdib_b;
	RGBQUAD *palette;
	unsigned char *dib_bits;
	int has_palette;
	int dib_pal_entries, tiff_pal_entries;
	int i,j;
	FILE *f;
	unsigned int num_tags;
	unsigned int sstring_offset, sstring_size,sstring_padding;
	unsigned int tiffbits_offset, tiffbits_size;
	unsigned int tiffpal_offset,  tiffpal_size;
	unsigned int tiffres_offset;
	unsigned int tiffbps_offset, tiffbps_size;
	int dibbytesperrow;
	int tiffbytesperrow;
	int pos;
	unsigned char *dib_row;
	unsigned char *temp_row;
	int promote1to4;

	int is_bw;  // 1bpp black&white
	int min_is_white;
	int tiff_bpp;

	TCHAR fn[MAX_PATH];

	temp_row = NULL;


	if(iinfo->img_fmt==FMT_TIFF)
		StringCbCopy(fn,sizeof(fn),iinfo->img_filename);
	else
		StringCbCopy(fn,sizeof(fn),_T("")); 	

	if(!pv_save_as_dialog(hwnd,_T("Save As TIFF"),fn,
		_T("TIFF image (*.tif)\0*.tif\0\0"),_T("tif")))
	{
		return 0;
	}


	lpdib = iinfo->lpdib;
	lpdib_b = (unsigned char*)iinfo->lpdib;

	if(lpdib->biSize < 40) {
		pv_mesg(hwnd,MSG_S,_T("Can't handle this type of BMP")); return 0;
	}
	if(lpdib->biCompression != BI_RGB) {
		pv_mesg(hwnd,MSG_S,_T("Can't handle this type of BMP")); return 0;
	}

	if(lpdib->biBitCount!=1 && lpdib->biBitCount!=4 &&
		lpdib->biBitCount!=8 && lpdib->biBitCount!=24)
	{
		pv_mesg(hwnd,MSG_S,_T("Can't handle this type of BMP")); return 0;
	}

	// some default settings that may get changed:
	tiff_bpp = lpdib->biBitCount;
	is_bw = 0;
	promote1to4=0;

	if(lpdib->biBitCount<=8) {
		has_palette = 1;
		palette = (RGBQUAD*) & lpdib_b[lpdib->biSize];
		dib_pal_entries=lpdib->biClrUsed;
		if(dib_pal_entries<1) {
			dib_pal_entries = 1<<lpdib->biBitCount;
		}

		// detect black & white images
		if(lpdib->biBitCount == 1) {
			tiff_bpp = 4; 
			if(palette[0].rgbRed==0 && palette[0].rgbGreen==0 && palette[0].rgbBlue==0
				&& (dib_pal_entries==1 || 
				(palette[1].rgbRed==255 && palette[1].rgbGreen==255 && palette[1].rgbBlue==255)))
			{
				is_bw=1; min_is_white=0;
				tiff_bpp = 1;
			}

			if(palette[0].rgbRed==255 && palette[0].rgbGreen==255 && palette[0].rgbBlue==255
				&& (dib_pal_entries==1 || 
				(palette[1].rgbRed==0 && palette[1].rgbGreen==0 && palette[1].rgbBlue==0)))
			{
				is_bw=1; min_is_white=1;
				tiff_bpp = 1;
			}

			if(is_bw) {
				tiff_bpp=1;
			}
			else {
				// ...... though most viewers support them, 
				// 1-bpp palette images apparently aren't baseline TIFF..
				// Guess we'll promote them to 4-bpp.
				tiff_bpp=4;
				promote1to4=1;
			}

		}

		if(is_bw) {
			has_palette = 0;
			tiff_pal_entries=0;
		}
		else {
			tiff_pal_entries= 1<<tiff_bpp;
		}
	}
	else {
		has_palette = 0;
		palette = NULL;
		dib_pal_entries=0;
		tiff_pal_entries=0;
	}

	//dib_bits = &lpdib_b[lpdib->biSize];
	dib_bits = &lpdib_b[iinfo->lpdib_bits_offs];

	dibbytesperrow = pv_get_dib_bytes_per_row(lpdib->biBitCount,lpdib->biWidth);

	tiffbytesperrow = (tiff_bpp * lpdib->biWidth + 7)/8;


	f=_tfopen(fn,_T("wb"));
	if(!f) {
		pv_mesg(hwnd,MSG_E,_T("Can't write file"));
		return 0;
	}

	write_bytes(f,little_endian?"II":"MM",2);
	write_uint16(f,42);
	write_uint32(f,8);  // first ifd offset

	// Calculate a bunch of file positions in advance, so
	// we can write the TIFF fields immediately without having
	// to somehow go back and modify them later. This is not robust,
	// but suffices for our simplistic purposes.

	num_tags = 13;
	if(has_palette) num_tags++;


	// data in the "oversize" area:
	//   xres, yres:    8+8=16
	//   bitspersample: 0 or 6
	//   palette        0 or 6x
	//   software_string  varies
	
	pos = 8 + 2 + 12*num_tags + 4;

	tiffres_offset = pos;
	pos+= 8+8;  // size of two RATIONALs

	tiffbps_offset = pos;
	tiffbps_size = (tiff_bpp>8)?6:0;
	pos+= tiffbps_size;

	tiffpal_offset = pos;
	tiffpal_size = tiff_pal_entries * 6;
	pos+= tiffpal_size;

	sstring_offset = pos;
	sstring_size = 1+strlen(einfo->software_name_full_a);
	sstring_padding = (sstring_size%2)?1:0;
	pos+= sstring_size+sstring_padding;

	tiffbits_offset = pos;

	tiffbits_size = tiffbytesperrow * lpdib->biHeight;

	// write the IFD
	write_uint16(f,num_tags);


	//**** tag: width
	write_uint16(f,256);
	write_uint16(f,TIFF_SHORT); // field type
	write_uint32(f,1); // count
	write_uint16(f,lpdib->biWidth); // value (or offset)
	write_uint16(f,0); // padding

	//**** tag: height
	write_uint16(f,257);
	write_uint16(f,TIFF_SHORT);
	write_uint32(f,1);
	write_uint16(f,lpdib->biHeight);
	write_uint16(f,0); // padding
	
	//**** tag: bitspersample
	write_uint16(f,258);
	write_uint16(f,TIFF_SHORT);
	if(tiff_bpp<=8) {
		write_uint32(f,1);
		write_uint16(f,tiff_bpp);
		write_uint16(f,0); // padding
	}
	else {
		write_uint32(f,3);
		write_uint32(f,tiffbps_offset);
	}

	//**** tag: compression
	write_uint16(f,259);
	write_uint16(f,TIFF_SHORT);
	write_uint32(f,1);
	write_uint16(f,1); // 1=no compression
	write_uint16(f,0); // padding

	//**** tag: photometric
	write_uint16(f,262);
	write_uint16(f,TIFF_SHORT);
	write_uint32(f,1);
	if(is_bw) {
		write_uint16(f,min_is_white?0:1);
	}
	else {
		write_uint16(f,(tiff_bpp<=8)?3:2); // 3=palette, 2=rgb
	}
	write_uint16(f,0); // padding

	//**** tag: stripoffsets
	write_uint16(f,273);
	write_uint16(f,TIFF_LONG);
	write_uint32(f,1);
	write_uint32(f,tiffbits_offset);


	//**** tag: samplesperpixel
	write_uint16(f,277);
	write_uint16(f,TIFF_SHORT);
	write_uint32(f,1);
	write_uint16(f,(tiff_bpp<=8)?1:3);
	write_uint16(f,0); // padding

	//**** tag: rowsperstrip
	write_uint16(f,278);
	write_uint16(f,TIFF_SHORT);
	write_uint32(f,1);
	write_uint16(f,lpdib->biHeight);
	write_uint16(f,0); // padding

	//**** tag: stripbytecounts
	write_uint16(f,279);
	write_uint16(f,TIFF_LONG);
	write_uint32(f,1);
	write_uint32(f,tiffbits_size);


	//**** tag: xresolution
	write_uint16(f,282);
	write_uint16(f,TIFF_RATIONAL);
	write_uint32(f,1);
	write_uint32(f,tiffres_offset);
	//**** tag: yresolution
	write_uint16(f,283);
	write_uint16(f,TIFF_RATIONAL);
	write_uint32(f,1);
	write_uint32(f,tiffres_offset+8);
	//**** tag: res units
	write_uint16(f,296);
	write_uint16(f,TIFF_SHORT);
	write_uint32(f,1);
	write_uint16(f,(iinfo->res_flag==1)?3:1);  // 1=unknown units, 3=centimeters
	write_uint16(f,0); // padding

	//**** tag: software
	write_uint16(f,305);
	write_uint16(f,TIFF_ASCII);
	write_uint32(f,sstring_size);
	write_uint32(f,sstring_offset);


	//**** tag: colormap
	if(has_palette) {
		write_uint16(f,320);
		write_uint16(f,TIFF_SHORT);
		write_uint32(f,tiff_pal_entries*3);
		write_uint32(f,tiffpal_offset);
	}

	
	//*************** done with tags
	
	write_uint32(f,0);  // next IFD (0=none)

	// write oversize fields

	// resolution
	if(iinfo->res_flag==1) { // meters
		// (divide meters by 100 to get centimeters)
		write_uint32(f,iinfo->res_x); write_uint32(f,100);
		write_uint32(f,iinfo->res_y); write_uint32(f,100);
	}
	else if(iinfo->res_flag==2) { // unknown units
		write_uint32(f,iinfo->res_x); write_uint32(f,1);
		write_uint32(f,iinfo->res_y); write_uint32(f,1);
	}
	else {   // resolution unknown
		write_uint32(f,1); write_uint32(f,1);
		write_uint32(f,1); write_uint32(f,1);
	}


	// bitspersample
	if(tiffbps_size>0) {
		write_uint16(f,8);
		write_uint16(f,8);
		write_uint16(f,8);
	}

	if(has_palette) {
		for(i=0;i<tiff_pal_entries;i++) {
			if(i<dib_pal_entries)
				write_uint16(f, (palette[i].rgbRed<<8)|palette[i].rgbRed );
			else
				write_uint16(f,0);
		}
		for(i=0;i<tiff_pal_entries;i++) {
			if(i<dib_pal_entries)
				write_uint16(f, (palette[i].rgbGreen<<8)|palette[i].rgbGreen );
			else
				write_uint16(f,0);
		}
		for(i=0;i<tiff_pal_entries;i++) {
			if(i<dib_pal_entries)
				write_uint16(f, (palette[i].rgbBlue<<8)|palette[i].rgbBlue );
			else
				write_uint16(f,0);
		}
	}

	write_bytes(f,einfo->software_name_full_a,sstring_size);
	if(sstring_padding>0) {
		write_uint8(f,0);
	}

	// write image data

	temp_row = (unsigned char*)malloc(tiffbytesperrow);
	if(!temp_row) {
		goto abort;
	}

	for(j=0;j<lpdib->biHeight;j++) {
		dib_row = &dib_bits[(lpdib->biHeight-1-j)*dibbytesperrow];

		if(lpdib->biBitCount==24) {
			// need to swap red & blue
			for(i=0;i<lpdib->biWidth;i++) {
				temp_row[i*3  ]=dib_row[i*3+2];
				temp_row[i*3+1]=dib_row[i*3+1];
				temp_row[i*3+2]=dib_row[i*3];
			}
			write_bytes(f, temp_row, tiffbytesperrow);

		}
		else if(promote1to4) {
			memset(temp_row,0,tiffbytesperrow);
			for(i=0;i<lpdib->biWidth;i++) {
				if( dib_row[i/8] & (1<<(7-i%8)) ) {
					temp_row[i/2] |= (i%2)?1:16;
				}
			}
			write_bytes(f, temp_row, tiffbytesperrow);
		}
		else {
			write_bytes(f, dib_row, tiffbytesperrow);
		}
	}

abort:
	if(temp_row) free(temp_row);
	fclose(f);

	return 1;
}
